# Helper Functions.

def parse_key(k)
  if /^[[:xdigit:]]{14}$/.match(k).nil?
    return NIL
  end  
  res = Array.new(7, 0)
  for i in 0..6
    res[i] = 1 + Integer("0x" + k[2*i,2])
  end
  return res
end

$invt = Array.new(257, 0)
def precompute()
  for i in 1..256
    for j in 1..256
      if i*j % 257 == 1
        $invt[i]=j
      end
    end
  end
end

def encrypt(s, k)
  a = k[4]
  b = k[5]
  last = k[6]
  encrypted = ""
  count = 0
  s.bytes{|c| 
    raise ("Error while encrypting. No null characters supported.") unless c != 0
    perm = [a,b,c].permutation(3).to_a[last % 6]
    last = (k[0] + perm[0]*k[1] + perm[1]*k[2] + perm[2]*k[3]) % 257
    if last == 256
      encrypted += "XX"
    else
      encrypted += "%02x" % last
    end
    count += 2
    if count % 80 == 0
      encrypted += "\n"
    end
    a = b
    b = c
  }
  return encrypted
end

def decrypt(s, k)
  s = s.gsub(/[[:space:]]/, "")
  decrypted = ""
  a = k[4]
  b = k[5]
  last = k[6]
  len = s.length
  raise ("Invalid length: %d" % len) unless len%2 == 0
  for i in 0..(len/2 - 1)
    if s[2*i, 2] == "XX"
      val = 256
    else
      val = Integer("0x" + s[2*i, 2])
    end
    perm = [a,b,0].permutation(3).to_a[last % 6]
    dec = (257*257*5 + val - k[0] - perm[0]*k[1] - perm[1]*k[2] - perm[2]*k[3])%257
    raise ("Error while decrypting.") unless dec != 0
    for i in 0..2
      if perm[i] == 0
        dec *= $invt[k[i+1]]
        dec %= 257
      end
    end
    last = val
    decrypted += "%c" % dec
    a = b
    b = dec
  end
  return decrypted
end

# Main Program.
if ARGV.length != 3
    print "Usage: ruby ", $PROGRAM_NAME, " key textfile command.\n"
    exit(1)
elsif parse_key(ARGV[0]).nil?
    print "Key ", ARGV[0], " invalid. Has to be a hexadecimal characters (0-9a-fA-F) of length 14 with.\n"
    exit(1)
elsif not File.exist?(ARGV[1])
    print "File ", ARGV[1], " does not exist.\n"
    exit(1)
elsif ARGV[2] != "encrypt" and ARGV[2] != "decrypt"
    print "Command ", ARGV[2], " not one of 'encrypt', 'decrypt'.\n"
    exit(1)
end

precompute()
s = ""
File.open(ARGV[1], "r").each_line do |l|
  s += l
end

if ARGV[2] == "encrypt"
  print encrypt(s, parse_key(ARGV[0])), "\n"
else
  print decrypt(s, parse_key(ARGV[0])), "\n"
end
